------------------------ MODULE MutualExclusionSpec ------------------------
\* Module MutualExclusionSpec provides general specification for mutual-exclusion problem.

\* Procs is set of processes.
\* Every process is assumed to loop and enter into "cs" state on every interation.
\* Non-critical state is represented as "non-cs".
\* When process decides it want to enter into critical-section, it first goes into "csentry" state.
CONSTANT Procs
VARIABLES pc

TypeOK == pc \in [Procs -> {"non-cs", "csentry", "cs"}]

vars == <<pc>>

\* All processes start from non-critical section.
Init == pc \in [Procs -> {"non-cs"}]

\* WantCS(proc) is action when proc decides that it wants to enter into critical section.
WantCS(proc) ==
        /\ pc[proc] = "non-cs"
        /\ pc' = [pc EXCEPT ![proc] = "csentry"]

\* EnterCS(proc) is action when proc enters critical section.
EnterCS(proc) ==
        /\ pc[proc] = "csentry"
        /\ \A i \in Procs \ {proc} : pc[i] /= "cs"
        /\ pc' = [pc EXCEPT ![proc] = "cs"]
        
\* ExitCS(proc) is action when proc leaves critical section.
ExitCS(proc) ==
        /\ pc[proc] = "cs"
        /\ pc' = [pc EXCEPT ![proc] = "non-cs"]

Next == \E i \in Procs: WantCS(i) \/ EnterCS(i) \/ ExitCS(i)


Spec == /\ Init
        /\ [][Next]_vars
        /\ \A i \in Procs: (            \* fairness
            \* not fair for WantCS - it can pause and even hang there
            /\ SF_vars(EnterCS(i))
            /\ SF_vars(ExitCS(i)))

----------------

\* MutualExclusion is invariant indicating that no two processes can be inside critical section at the same time.
MutualExclusion == \A i,j \in Procs: (i /= j) => ~((pc[i] = "cs") /\ (pc[j] = "cs"))

\* Liveness is temporal property indicating that every process has a chance to enter critical section and leaves it.
\* A process should keep on entering the critical-section unless it is stuck in its non-critical part.
Liveness == \A i \in Procs:
                /\
                    \/ <>[](pc[i] = "non-cs")                   \* either stuck in non-cs, or
                    \/ []<>(pc[i] = "csentry")                  \* keeps on reaching "enter cs"
                /\
                    /\ (pc[i] = "csentry") ~> (pc[i] = "cs")    \* if it wants to enter, it enters cs and leaves it
                    /\ (pc[i] = "cs") ~> (pc[i] = "non-cs")


THEOREM Spec => []TypeOK        
THEOREM Spec => []MutualExclusion
THEOREM Spec => Liveness

=============================================================================
