(*
-------- CONFIG Route ---------

CONSTANT NoRoute = ZZZNoRoute
CONSTANT Inf = ZZZInf

=================
*)

------------------------------- MODULE Route -------------------------------
EXTENDS Integers, FiniteSets, Sequences

CONSTANT Nodes, Edges, root

\* Has2Items(e) is true when e is set with exactly 2 elements.
\*Has2Items(s) == \E a, b \in s : (a /= b) /\ (s = {a,b})
\*Has2Items(s) == Cardinality(s) = 2


ASSUME /\ root \in Nodes
       /\ \A e \in Edges : (e \subseteq Nodes) /\ (Cardinality(e) = 2)


\* Connected is true if two nodes i and j are interconnected by some path.
RECURSIVE _Connected(_,_,_)
_Connected(i,j,Intra) ==
        CASE i = j           -> TRUE
          [] {i,j} \in Edges -> TRUE
          [] Intra = {}      -> FALSE
          [] OTHER           -> \E x \in Intra : ({i,x} \in Edges) /\ _Connected(x,j, Intra\{x})

Connected(i,j) == _Connected(i,j, Nodes\{i,j})

\* All nodes must be connected
ASSUME \A i,j \in Nodes : Connected(i,j)



\* ReachableFrom returns set of nodes reachable from nodes in S.
RECURSIVE ReachableFrom(_)
ReachableFrom(S) ==
        LET S_ == S \cup {n \in (Nodes \ S) : \E x \in S : {x,n} \in Edges}
        IN
            IF S_ = S THEN S
                      ELSE ReachableFrom(S_)

ASSUME ReachableFrom({root}) = Nodes


--------

\* < on Nat \cup {∞}
CONSTANT Inf    \* = Inf
m \prec n  ==  IF m = Inf THEN FALSE
                          ELSE IF n = Inf THEN TRUE ELSE m < n

CONSTANT NoRoute \* = NoRoute


\* Neigh(n) is set of n's neighbours.
Neigh(n) == {x \in Nodes : {x,n} \in Edges}

\* EdgeQQ({n,m}) is {<n,m>, <m,n>}
EdgeQQ(e) == LET x == CHOOSE x \in e : TRUE
             IN  LET y == CHOOSE y \in (e\{x}) : TRUE
                 IN
                    {<<x,y>>, <<y,x>>}

\* All n->m and m<-n queues
Queues == {e \in Nodes \X Nodes : {e[1],e[2]} \in Edges}

QueuesFrom(n) == {q \in Queues : q[1] = n}
QueuesTo(n)   == {q \in Queues : q[2] = n}


(********
--algorithm Route {
  variables
    \* msgs[<<n,m>>] is message queue for n->m
\*  msgs = [q \in (UNION {EdgeQQ(e) : e \in Edges})  |->
    msgs = [q \in Queues  |->
                   IF q[1] = root THEN <<0>>  \* start with root sending initial messages
                                  ELSE << >>];

  fair process (RT \in Nodes)
    variables depth  =  IF self = root THEN 0 ELSE Inf,
              parent =  IF self = root THEN root ELSE NoRoute;
  {
    loop: while (TRUE) {
            with (q \in {r \in QueuesTo(self) : msgs[r] /= << >>}) {
              with (m = Head(msgs[q])) {
              with (msgs_ = [msgs EXCEPT ![q] = Tail(msgs[q])]) {
                if (depth = Inf \/ m < depth - 1) {
                  depth := m + 1;
                  parent := q[1];
                  \* send m further to all neighbours except from whom we received it
                  with (msgs__ = [t \in Queues |->
                            IF t \in QueuesFrom(self) \ {<<self, q[1]>>}  \* XXX not q[1]->parent would be incorrect
                                THEN Append(msgs_[t], depth)
                                ELSE msgs_[t] ])
                    msgs := msgs__;
                }
                else {
                  msgs := msgs_
                }
              }}
            }
          }
  }
};
********)
\* BEGIN TRANSLATION (chksum(pcal) = "b56d10df" /\ chksum(tla) = "c40e6a89")
VARIABLES msgs, depth, parent

vars == << msgs, depth, parent >>

ProcSet == (Nodes)

Init == (* Global variables *)
        /\ msgs = [q \in Queues  |->
                          IF q[1] = root THEN <<0>>
                                         ELSE << >>]
        (* Process RT *)
        /\ depth = [self \in Nodes |-> IF self = root THEN 0 ELSE Inf]
        /\ parent = [self \in Nodes |-> IF self = root THEN root ELSE NoRoute]

RT(self) == \E q \in {r \in QueuesTo(self) : msgs[r] /= << >>}:
              LET m == Head(msgs[q]) IN
                LET msgs_ == [msgs EXCEPT ![q] = Tail(msgs[q])] IN
                  IF depth[self] = Inf \/ m < depth[self] - 1
                     THEN /\ depth' = [depth EXCEPT ![self] = m + 1]
                          /\ parent' = [parent EXCEPT ![self] = q[1]]
                          /\ LET msgs__ ==      [t \in Queues |->
                                           IF t \in QueuesFrom(self) \ {<<self, q[1]>>}
                                               THEN Append(msgs_[t], depth'[self])
                                               ELSE msgs_[t] ] IN
                               msgs' = msgs__
                     ELSE /\ msgs' = msgs_
                          /\ UNCHANGED << depth, parent >>

Next == (\E self \in Nodes: RT(self))

Spec == /\ Init /\ [][Next]_vars
        /\ \A self \in Nodes : WF_vars(RT(self))

\* END TRANSLATION


Terminated == \A q \in Queues : msgs[q] = << >>

\* SeqDecreasing is true if S is sequence with decreasing elements
RECURSIVE SeqDecreasing(_)
SeqDecreasing(S) ==
        CASE S = << >>  -> TRUE
          [] Len(S) = 1 -> TRUE
          [] OTHER      -> (S[2] < S[1])  /\  SeqDecreasing(Tail(S))

\* SeqSet converts sequence S to set of its values.
SeqSet(S) == {S[x]: x \in DOMAIN S}

TypeOK ==
        /\ depth \in [Nodes -> 0..Cardinality(Nodes) \cup {Inf}]
        /\ parent \in [Nodes -> Nodes \cup {NoRoute}]
        /\ msgs \in [Queues -> Seq(0..(Cardinality(Nodes)-1))]
        /\ \A q \in Queues :
                /\ SeqDecreasing(msgs[q])
                /\ \A d \in SeqSet(msgs[q]) : d >= depth[q[1]]


\* ParentPath returns path from node n to root constructed via parents.
\* The path is of the form << n, ..., root | NoRoute >>.
RECURSIVE ParentPath(_)
ParentPath(n) ==
    CASE n = root               -> << root >>
      [] parent[n] = NoRoute    -> << n, NoRoute >>
      [] OTHER                  -> << n >> \o ParentPath(parent[n])


\* PathCorrect is true if p is correct path going through Edges.
RECURSIVE PathCorrect(_)
PathCorrect(p) ==
    CASE p = << >>   -> FALSE
      [] Len(p) = 1  -> p[1] \in Nodes
      [] OTHER       -> {p[1],p[2]} \in Edges /\ PathCorrect(Tail(p))

\* SeqN is like Seq but produces sequences with Len =< N
SeqN(S, N) == UNION {[1..n -> S] : n \in 1..N}


\* PathMinimal verifies that path p has minimal length in between p0 and pEND.
PathMinimal(p) == ~ \E pp \in SeqN(Nodes, Cardinality(Nodes)) :
                        /\ pp /= << >>
                        /\ Len(pp) < Len(p)
                        /\ pp[1] = p[1]
                        /\ pp[Len(pp)] = p[Len(p)]
                        /\ PathCorrect(pp)


\* Dist is len of minimal path in between n-m.
RECURSIVE _Dist(_, _)
_Dist(n, S) ==
    IF n \in S THEN 1
    ELSE
        LET S_ == S \cup {x \in Nodes : \E y \in S : {x,y} \in Edges}
        IN
            IF S_ = S THEN Inf
                      ELSE
                           LET d == _Dist(n, S_)
                           IN
                                IF d = Inf THEN Inf
                                           ELSE 1 + d
Dist(n,m) == _Dist(n, {m})

\* Paths are OK if they are all correct and minimal.
PathsOK == \A n \in Nodes :
                 LET p == ParentPath(n)
                 IN
                     /\ p[1] = n
                     /\ p[Len(p)] = root  \* not NoRoute
                     /\ PathCorrect(p)
                     /\ PathMinimal(p)
                     /\ Len(p) = Dist(n, root)

\* Paths must be all OK when terminated.
THEOREM Terminated => PathsOK


Inv == \A n \in Nodes :
            /\ depth[n] = Inf \/ depth[n]+1 >= Dist(n, root)
            /\ (depth[n] /= Inf) => (LET p == ParentPath(n) IN
                                        /\ p[Len(p)] = root
                                        /\ Len(p) =< depth[n]+1)

--------

RECURSIVE SumSeq(_)
SumSeq(S) == IF S = << >>  THEN 0
                           ELSE Head(S) + SumSeq(Tail(S))


RECURSIVE SumSet(_)
SumSet(S) == IF S = {}  THEN 0
                        ELSE
                            LET y == CHOOSE x \in S : TRUE
                            IN y + SumSet(S \ {y})



------

Eventually_Terminated   == <>Terminated
Terminated_then_PathsOK == Terminated => PathsOK

=============================================================================
