#!/usr/bin/env python
"""Verify Route.tla on all connected graphs up to N nodes"""

from __future__ import print_function
import sys
from subprocess import check_call


class Graph:
    # .nodes    {nodes}
    # .edges    {edges}

    def __init__(g, nodes, edges):
        g.nodes = set(nodes)
        g.edges = set(edges)

    def __str__(g):
        nv = ['%s' % _ for _ in g.nodes]
        ev = [setstr(_) for _ in g.edges]
        return "Graph nodes={%s}  edges={%s}" % (','.join(nv), ', '.join(ev))


def setstr(s):
    assert isinstance(s, (set,frozenset)), s
    v = ['%s' % xstr(_) for _ in s]
    return '{%s}' % ','.join(v)

def xstr(obj):
    if isinstance(obj, (set,frozenset)):
        return setstr(obj)
    return str(obj)


# generate all connected graphs with N(nodes) =< N
def genGraphs(N):
    assert N >= 1

    nodes  = []         # of all nodes
    Sprevedges = set()  # set of edges for all graphs with (i-1) nodes

    for i in range(1,N+1):
        print()
        n = ('r' if i == 1 else 'n%d' % (i-1))
        nodes = nodes + [n]

        # only 1 node
        if i == 1:
            yield Graph(nodes, set())
            continue

        # only 2 nodes
        if i == 2:
            yield Graph(nodes, set([frozenset(nodes),]))
            Sprevedges.add(frozenset([frozenset(nodes)]))
            continue


        # induction: the graph remains connected if previous (N-1) graph is
        # connected, and n is connected to at least one node in the previous graph.
        #
        # FIXME wrong: (N-1) graph might be not connected and if n is adding
        # connection to enough nodes N-graph might become connected. Below
        # approach thus skips some possible connected graphs.
        Sedges = set()

        # for all combinations to which nodes n is connected in the previous graph.
        for j in range(1, 1<<(i-1)):
            newconn = set()
            for k in range(i-1):
                if j & (1 << k):
                    newconn.add(frozenset([nodes[k], n]))

            #print("j: %d\tnewconn: %s" % (j, newconn))
            #print("Sprevedges: %s" % Sprevedges)

            for edges in Sprevedges:
                newedges = edges.union(newconn)
                #print("edges:    %s" % edges)
                #print("newconn:  %s" % newconn)
                #print("newedges: %s" % newedges)
                Sedges.add(newedges)
                yield Graph(nodes, newedges)

        Sprevedges = Sedges

# verify Route.tla on graph G.
_cfg_base = """
SPECIFICATION
Spec

INVARIANT
TypeOK
Inv
Terminated_then_PathsOK

PROPERTY
Eventually_Terminated

CONSTANT NoRoute = NoRoute
CONSTANT Inf = Inf
"""
def TLCRoute(G):
    cfg = _cfg_base
    cfg += "\n\n\nCONSTANTS\n"
    for n in G.nodes:
        cfg += "%s = %s\n" % (n, n)
    cfg += "\n"
    cfg += "CONSTANT root = r\n"
    cfg += "CONSTANT Nodes = %s\n" % setstr(G.nodes)
    cfg += "CONSTANT Edges = %s\n" % setstr(G.edges)

    writefile("Route.cfg", cfg)

    check_call(["./tlc", "-workers", "auto", "-deadlock", "Route.tla"])




def writefile(path, data):
    with open(path, "w") as f:
        f.write(data)



def main():
    N = int(sys.argv[1])
    for g in genGraphs(N):
        print()
        print(g)
        TLCRoute(g)



if __name__ == '__main__':
    main()
